from dataclasses import dataclass
from difflib import SequenceMatcher
from pathlib import Path
from typing import List, Optional, Tuple
import argparse

@dataclass
class Args:
    input_file: Path
    output_file: Path
    verbose: bool
    marked: bool

@dataclass
class LineInfo:
    content: str  # Original content for comparison
    display: str  # Content for display/output (may include markers)

def find_longest_common_substring(s1: str, s2: str) -> Tuple[int, str]:
    """Returns length and content of longest common substring between two strings."""
    matcher = SequenceMatcher(None, s1, s2)
    match = matcher.find_longest_match(0, len(s1), 0, len(s2))
    return match.size, s1[match.a:match.a + match.size]

def find_most_similar_line(target: str, lines: List[str]) -> Tuple[int, str]:
    """Returns index and content of the most similar line to target."""
    if not lines:
        raise ValueError("Empty lines list")
    
    similarities = [
        (i, find_longest_common_substring(target, line)[0])
        for i, line in enumerate(lines)
    ]
    most_similar_idx = max(similarities, key=lambda x: x[1])[0]
    return most_similar_idx, lines[most_similar_idx]

def sort_by_similarity(lines: List[str], verbose: bool = False, marked: bool = False) -> List[str]:
    if not lines:
        return []
    
    result: List[str] = []
    working_lines = lines.copy()
    first_line: str = working_lines.pop(0)
    current = LineInfo(first_line, first_line)
    
    while working_lines:
        if verbose:
            print(current.display)
        result.append(current.display)
        
        idx, next_line = find_most_similar_line(current.content, working_lines)
        next_content = working_lines[idx]
        next_display = next_line
        
        if marked:
            _, common = find_longest_common_substring(current.content, next_content)
            next_display = next_line.replace(common, f"[{common}]")
        
        working_lines.pop(idx)
        current = LineInfo(next_content, next_display)
    
    result.append(current.display)
    if verbose:
        print(current.display)
    
    return result

def main(args: Args) -> None:
    lines = args.input_file.read_text(encoding='utf-8').splitlines()
    sorted_lines = sort_by_similarity(lines, args.verbose, args.marked)
    args.output_file.write_text('\n'.join(sorted_lines), encoding='utf-8')

def parse_args() -> Args:
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=Path)
    parser.add_argument('output_file', type=Path)
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--marked', action='store_true')
    args = parser.parse_args()
    return Args(args.input_file, args.output_file, args.verbose, args.marked)

if __name__ == '__main__':
    main(parse_args())
