#!/usr/bin/env python3
from typing import List, Set, Tuple
from pathlib import Path
import argparse
from difflib import SequenceMatcher

MIN_LENGTH: int = 1

def get_words(line: str) -> Set[str]:
    """Extract words longer than 2 chars from line."""
    return {word for word in line.split() if len(word) > MIN_LENGTH - 1}

def find_common_words(line1: str, line2: str) -> Set[str]:
    """Find common words between two lines."""
    words1 = get_words(line1)
    words2 = get_words(line2)
    return words1 & words2

def mark_common_words(line: str, common: Set[str]) -> str:
    """Mark common words with square brackets."""
    result = []
    for word in line.split():
        if word in common and len(word) > MIN_LENGTH - 1:
            result.append(f"[{word}]")
        else:
            result.append(word)
    return " ".join(result)

def find_most_similar(current: str, lines: List[str]) -> Tuple[str, int]:
    """Find line most similar to current based on common word count."""
    max_common = -1
    most_similar = ""
    
    for line in lines:
        common = find_common_words(current, line)
        if len(common) > max_common:
            max_common = len(common)
            most_similar = line
    
    return most_similar, max_common

def sort_lines(lines: List[str], verbose: bool = False, marked: bool = False) -> List[str]:
    """Sort lines by similarity based on common words."""
    result: List[str] = []
    working_lines = lines.copy()
    
    if not working_lines:
        return result
        
    current = working_lines[0]
    working_lines.remove(current)
    
    if verbose:
        print(f"Starting with: {current}")
    result.append(current)
    
    while working_lines:
        next_line, common_count = find_most_similar(current, working_lines)
        working_lines.remove(next_line)
        
        if marked and result:  # Don't mark first line
            common_words = find_common_words(current, next_line)
            marked_line = mark_common_words(next_line, common_words)
            result.append(marked_line)
            if verbose:
                print(f"Adding (common words: {common_count}): {marked_line}")
        else:
            result.append(next_line)
            if verbose:
                print(f"Adding (common words: {common_count}): {next_line}")
                
        current = next_line
        
    return result

def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--marked', action='store_true')
    args = parser.parse_args()

    input_path = Path("pan_tadeusz_znormalizowany.txt")
    output_path = Path("pan_tadeusz_posortowany_wedlug_wspolnych_slow.txt")
    
    assert input_path.exists(), f"Input file {input_path} not found"
    
    with open(input_path, encoding='utf-8') as f:
        lines = [line.strip() for line in f if line.strip()]
        
    sorted_lines = sort_lines(lines, args.verbose, args.marked)
    
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write('\n'.join(sorted_lines))

# Tests
def test_get_words() -> None:
    assert get_words("a the test string") == {"the", "test", "string"}
    assert get_words("") == set()
    assert get_words("a b c") == set()

def test_find_common_words() -> None:
    line1 = "the quick brown fox"
    line2 = "the lazy brown dog"
    assert find_common_words(line1, line2) == {"the", "brown"}

def test_mark_common_words() -> None:
    line = "the quick brown fox jumps"
    common = {"brown", "jumps"}
    assert mark_common_words(line, common) == "the quick [brown] fox [jumps]"

if __name__ == "__main__":
    # testy zakładają MIN_LENGTH = 3
    #test_get_words()
    #test_find_common_words()
    #test_mark_common_words()
    main()
