import argparse
from pathlib import Path
from typing import List, Tuple, Optional
import sys

def get_common_prefix_length(str1: str, str2: str) -> int:
    """Returns the length of common prefix of two strings."""
    min_len = min(len(str1), len(str2))
    for i in range(min_len):
        if str1[i] != str2[i]:
            return i
    return min_len

def get_common_suffix_length(str1: str, str2: str) -> int:
    """Returns the length of common suffix of two strings."""
    return get_common_prefix_length(str1[::-1], str2[::-1])

def similarity_score(str1: str, str2: str) -> int:
    """Calculate similarity score based on common prefix and suffix lengths."""
    return get_common_prefix_length(str1, str2) + get_common_suffix_length(str1, str2) + (get_common_prefix_length(str1, str2) * get_common_suffix_length(str1, str2))

def mark_common_parts(current: str, previous: str) -> str:
    """
    Marks common parts (>=1 chars) between current and previous line with square brackets.
    Returns modified current line string.
    """
    prefix_len = get_common_prefix_length(current, previous)
    suffix_len = get_common_suffix_length(current, previous)
    
    result = current
    if prefix_len >= 1:
        result = f"[{current[:prefix_len]}]{current[prefix_len:]}"
    if suffix_len >= 1:
        # If we already have prefix marked, we need to handle the brackets carefully
        if prefix_len >= 1:
            base = result[prefix_len + 2:]  # Skip "[prefix]"
            suffix_start = len(base) - suffix_len
            result = f"[{current[:prefix_len]}]{base[:suffix_start]}[{base[suffix_start:]}]"
        else:
            suffix_start = len(result) - suffix_len
            result = f"{result[:suffix_start]}[{result[suffix_start:]}]"
    return result

def find_most_similar(target: str, candidates: List[str]) -> str:
    """
    Finds the most similar string to target from candidates list.
    Returns the found string.
    """
    assert candidates, "Candidates list cannot be empty"
    
    max_score = -1
    most_similar = candidates[0]
    
    for candidate in candidates:
        score = similarity_score(target, candidate)
        if score > max_score:
            max_score = score
            most_similar = candidate
            
    return most_similar

def sort_by_similarity(
    lines: List[str],
    verbose: bool = False,
    marked: bool = False
) -> List[str]:
    """
    Sorts lines by similarity using the specified algorithm.
    Returns list of sorted lines (with optional marking).
    """
    if not lines:
        return []
        
    working_lines = lines.copy()
    result: List[str] = []
    display_lines: List[str] = []
    
    current = working_lines[0]
    display_current = current
    
    while working_lines:
        if verbose:
            print(f"Adding: {display_current}")
            
        result.append(current)
        display_lines.append(display_current)
        working_lines.remove(current)
        
        if not working_lines:
            break
            
        current = find_most_similar(current, working_lines)
        display_current = current
        if marked and result:
            display_current = mark_common_parts(current, result[-1])
            
    return display_lines if marked else result

def main() -> None:
    parser = argparse.ArgumentParser(description='Sort text file lines by similarity')
    parser.add_argument('input_file', type=str, help='Input file path')
    parser.add_argument('output_file', type=str, help='Output file path')
    parser.add_argument('--verbose', action='store_true', help='Print progress')
    parser.add_argument('--marked', action='store_true', help='Mark common parts with []')
    
    args = parser.parse_args()
    
    try:
        with open(args.input_file, 'r', encoding='utf-8') as f:
            lines = [line.strip() for line in f.readlines() if line.strip()]
    except FileNotFoundError:
        print(f"Error: Input file '{args.input_file}' not found", file=sys.stderr)
        sys.exit(1)
    
    sorted_lines = sort_by_similarity(lines, args.verbose, args.marked)
    
    try:
        with open(args.output_file, 'w', encoding='utf-8') as f:
            for line in sorted_lines:
                f.write(f"{line}\n")
    except IOError as e:
        print(f"Error writing to output file: {e}", file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()
